from aioredis_models import RedisDoubleHash
from .redis_tests import RedisTests


class RedisDoubleHashTests(RedisTests):
    _key = 'test-key'
    _inverse_key = 'key-test'
    _redis_double_hash: RedisDoubleHash = None

    async def asyncSetUp(self):
        await super().asyncSetUp()

        self._redis_double_hash = RedisDoubleHash(self._redis, self._key, self._inverse_key)
        await self._redis_double_hash.delete()

    async def test_fields_gets_fields(self):
        await self._redis_double_hash.set('foo', 'bar')
        await self._redis_double_hash.set('foo', 'bat')
        await self._redis_double_hash.set('biz', 'boo')

        result = await self._redis_double_hash.fields()

        self.assertEqual(result, {'foo', 'biz'})

    async def test_fields_inverted_gets_inverted_fields(self):
        await self._redis_double_hash.set('foo', 'bar')
        await self._redis_double_hash.set('foo', 'bat')
        await self._redis_double_hash.set('biz', 'boo')

        result = await self._redis_double_hash.fields_inverted()

        self.assertEqual(result, {'bar', 'bat', 'boo'})

    async def test_set_sets_mapping(self):
        await self._redis_double_hash.set('foo', 'bar')
        await self._redis_double_hash.set('foo', 'bat')

        foo = await self._redis_double_hash.get('foo')
        bar = await self._redis_double_hash.get_inverted('bar')
        bat = await self._redis_double_hash.get_inverted('bat')

        self.assertEqual(set(foo), {'bar', 'bat'})
        self.assertEqual(bar, ['foo'])
        self.assertEqual(bat, ['foo'])

    async def test_set_inverted_sets_mapping(self):
        await self._redis_double_hash.set_inverted('foo', 'bar')
        await self._redis_double_hash.set_inverted('foo', 'bat')

        bar = await self._redis_double_hash.get('bar')
        bat = await self._redis_double_hash.get('bat')
        foo = await self._redis_double_hash.get_inverted('foo')

        self.assertEqual(set(foo), {'bar', 'bat'})
        self.assertEqual(bar, ['foo'])
        self.assertEqual(bat, ['foo'])

    async def test_unset_unsets_mapping(self):
        await self._redis_double_hash.set('foo', 'bar')
        await self._redis_double_hash.set('foo', 'bat')

        await self._redis_double_hash.unset('foo', 'bat')

        foo = await self._redis_double_hash.get('foo')
        bar = await self._redis_double_hash.get_inverted('bar')
        bat = await self._redis_double_hash.get_inverted('bat')

        self.assertEqual(foo, ['bar'])
        self.assertEqual(bar, ['foo'])
        self.assertEqual(bat, [])

    async def test_remove_removes_mapping(self):
        await self._redis_double_hash.set('foo', 'bar')
        await self._redis_double_hash.set('foo', 'bat')
        await self._redis_double_hash.set('bar', 'foo')

        await self._redis_double_hash.remove('foo')

        foo = await self._redis_double_hash.get('foo')
        bar_fwd = await self._redis_double_hash.get('bar')
        bar_bwd = await self._redis_double_hash.get_inverted('bar')
        bat = await self._redis_double_hash.get_inverted('bat')

        self.assertEqual(foo, [])
        self.assertEqual(bar_fwd, ['foo'])
        self.assertEqual(bar_bwd, [])
        self.assertEqual(bat, [])

    async def test_remove_inverted_removes_mapping(self):
        await self._redis_double_hash.set('foo', 'bar')
        await self._redis_double_hash.set('foo', 'bat')
        await self._redis_double_hash.set('biz', 'bat')

        await self._redis_double_hash.remove_inverted('bat')

        foo = await self._redis_double_hash.get('foo')
        biz = await self._redis_double_hash.get('biz')
        bar = await self._redis_double_hash.get_inverted('bar')
        bat = await self._redis_double_hash.get_inverted('bat')

        self.assertEqual(foo, ['bar'])
        self.assertEqual(biz, [])
        self.assertEqual(bar, ['foo'])
        self.assertEqual(bat, [])

    async def test_delete_deletes_all(self):
        await self._redis_double_hash.set('foo', 'bar')
        await self._redis_double_hash.set('foo', 'bat')

        await self._redis_double_hash.delete()

        foo = await self._redis_double_hash.get('foo')
        bar = await self._redis_double_hash.get_inverted('bar')
        bat = await self._redis_double_hash.get_inverted('bat')

        self.assertFalse(foo)
        self.assertFalse(bar)
        self.assertFalse(bat)



"""
This module contains the following classes:
- RedisModel: Base class for all Redis models.
"""

from typing import Union
from aioredis import Redis
from .redis_client import RedisClient


class RedisModel:
    """
    Base class for all Redis models.
    """

    def __init__(
        self,
        redis: Union[Redis, RedisClient]  # pylint:disable=unsubscriptable-object
    ):
        """
        Creates a new instance of `RedisModel`.

        Args:
            redis (Union[Redis, RedisClient]): The Redis instance to use to connect to Redis.
                This can be an instance of `RedisClient` to allow controlling transactions
                externally.
        """

        self._redis =  redis \
            if isinstance(redis, RedisClient) else RedisClient(redis)

        self.begin_transaction = self._redis.begin_transaction
        self.is_in_transaction = self._redis.is_in_transaction
        self.discard_transaction = self._redis.discard_transaction
        self.execute_transaction = self._redis.execute_transaction
        self.get_connection = self._redis.get_connection



"""
This module contains the following classes:
- RedisHash: Represents a hash map stored in Redis.
"""

from typing import List, Any, Awaitable, AsyncIterator
from .redis_key import RedisKey
from .asyncio_utils import noop


class RedisHash(RedisKey):
    """
    Represents a hash map stored in Redis.
    """

    def length(self) -> Awaitable[int]:
        """
        Gets the length of the hash map.

        Returns:
            Awaitable[int]: The length of the hash map.
        """

        return self.get_connection().hlen(self._key)

    def field_length(self, field: str) -> Awaitable[int]:
        """
        Gets the length of the given field.

        Args:
            field (str): The field to check.

        Returns:
            Awaitable[int]: The length of the given field.
        """

        return self.get_connection().hstrlen(self._key, field)

    def field_exists(self, field: str) -> Awaitable[bool]:
        """
        Checks whether the given field exists.

        Args:
            field (str): The field to check.

        Returns:
            Awaitable[bool]: Whether the field exists or not.
        """

        return self.get_connection().hexists(self._key, field)

    def fields(self, encoding='utf-8') -> Awaitable[List]:
        """
        Gets all the fields in the hash map.

        Args:
            encoding (str, optional): The encoding to use for decoding the field keys. Defaults to
                'utf-8'.

        Returns:
            Awaitable[List]: The set of fields in the hash map.
        """

        return self.get_connection().hkeys(self._key, encoding=encoding)

    def get_all(self, encoding='utf-8') -> Awaitable[dict]:
        """
        Gets the entire hash map.

        Args:
            encoding (str, optional): The encoding to use for decoding the keys and values.
                Defaults to 'utf-8'.

        Returns:
            Awaitable[dict]: The hash map.
        """

        return self.get_connection().hgetall(self._key, encoding=encoding)

    def get(self, field: str, encoding='utf-8') -> Awaitable[Any]:
        """
        Gets the value of the given field in the hash map.

        Args:
            field (str): The field to get.
            encoding (str, optional): The encoding to use for decoding the values. Defaults to
                'utf-8'.

        Returns:
            Awaitable[Any]: The value of the field.
        """

        return self.get_connection().hget(self._key, field, encoding=encoding)

    async def enumerate(
        self,
        field_pattern: str=None,
        batch_size: int=None,
        encoding: str='utf-8'
    )-> AsyncIterator[Any]:
        """
        Enumerates over the items of the hash using HSCAN command. This operation is not atomic and
        cannot be performed transactionally.

        Args:
            field_pattern (str, optional): A string to filter fields with, if needed.
                Defaults to None.
            batch_size (int, optional): The maximum number of items to get with each scan.
                Defaults to None.

        Returns:
            AsyncIterator[Any]: An iterator that can be used to iterate over the result.
        """
        cursor = '0'
        while cursor != 0:
            cursor, data = await self.get_connection().hscan(
                self._key, cursor=cursor, match=field_pattern, count=batch_size
            )
            for key, value in data:
                yield (key, value) if not encoding else (
                    key.decode(encoding),
                    value.decode(encoding)
                )

    def set_all(self, values: dict):
        """
        Sets the entire hash map to the given `dict`.

        Args:
            values (dict): A `dict` containing the key/value map to set.
        """

        if values:
            return self.get_connection().hmset_dict(self._key, values)
        return noop()

    def set(self, field: str, value: str):
        """
        Set the value of the given field.

        Args:
            field (str): The field whose value is to be set.
            value (str): The value to set for the given field.
        """

        if value:
            return self.get_connection().hset(self._key, field, value)
        return noop()

    def remove(self, field: str) -> Awaitable[int]:
        """
        Removes the given field from the hash map.

        Args:
            field (str): The field to remove.

        Returns:
            Awaitable[int]: The number of field removed from the hash map.
        """

        return self.get_connection().hdel(self._key, field)



"""
This module contains the following classes:
- RedisKey: represents a generic Redis key.
"""

from typing import Awaitable, Union
from aioredis import Redis
from .redis_model import RedisModel
from .redis_client import RedisClient


class RedisKey(RedisModel):
    """
    Represents a Redis key of any type. Acts as the class for all data structures.
    """

    def __init__(
        self,
        redis: Union[Redis, RedisClient],  # pylint:disable=unsubscriptable-object
        key: str
    ):
        """
        Creates an instance ot `RedisKey`.

        Args:
            redis (Union[Redis, RedisClient]): The Redis instance to use to connect to Redis.
                This can be an instance of `RedisClient` to allow controlling transactions
                externally.
            key (str): The Redis key to use.
        """

        super().__init__(redis)
        self._key = key

    def delete(self) -> Awaitable[int]:
        """
        Deletes the key from Redis.

        Returns:
            Awaitable[int]: The number of items that were deleted.
        """

        return self.get_connection().delete(self._key)

    async def exists(self) -> Awaitable[bool]:
        """
        Checks if the key exists in Redis or not. This operation cannot be performed
        transactionally.

        Returns:
            Awaitable[bool]: A flag indicating whether the key exists or not.
        """

        return await self.get_connection().exists(self._key) > 0



"""
This module contains the following classes:
- RedisClient: A class that wraps the Redis client from aioredis and simplifies support for
    transactions.
"""

from typing import Awaitable, List, Union
from aioredis import Redis
from aioredis.commands import MultiExec
from .redis_transaction import RedisTransaction


class RedisClient:
    """
    A class that wraps the Redis client from aioredis and simplifies support for transactions.
    """
    _multi_exec: MultiExec = None

    def __init__(self, redis: Redis):
        """
        Creates a new instance of `RedisClient`.

        Args:
            redis (Redis): The Redis instance to use to connect to Redis.
        """
        self._redis = redis

    def get_connection(self) -> Union[Redis, MultiExec]:  # pylint:disable=unsubscriptable-object
        """
        Gets the Redis connection currently in use.

        Returns:
            Union[Redis, MultiExec]: The Redis instance to use to connect to Redis. This can be
                a `MultiExec` instance if a transaction is in progress.
        """
        return self._multi_exec or self._redis

    def is_in_transaction(self) -> bool:
        """
        Returns a value indicating whether a transaction is in progress.

        Returns:
            bool: Whether a transaction is in progress.
        """
        return self._multi_exec is not None

    def begin_transaction(self) -> RedisTransaction:
        """
        Begins a transaction.

        Returns:
            RedisTransaction: A `RedisTransaction` instance that can be used to interact with the
                transaction.
        """
        assert not self.is_in_transaction()

        self._multi_exec = self._redis.multi_exec()
        return RedisTransaction(self)

    def discard_transaction(self):
        """
        Discards the current transaction. The client is responsible for canceling pending tasks.
        """

        assert self._multi_exec
        self._multi_exec = None

    def execute_transaction(self) -> Awaitable[List]:
        """
        Executes the transaction. The client is responsible for awaiting pending tasks.

        Returns:
            Awaitable[List]: The result of the transaction operations.
        """

        assert self._multi_exec

        result = self._multi_exec.execute()
        self._multi_exec = None
        return result



from typing import List
from aioredis_models import RedisList, RedisString, RedisKey, RedisClient
from .redis_tests import RedisTests


class RedisTransactionTests(RedisTests):
    _redis_keys: List[RedisKey] = []

    def add_redis_item(self, redis_key: RedisKey):
        self._redis_keys.append(redis_key)
        return redis_key

    async def asyncSetUp(self):
        await super().asyncSetUp()
        self._redis_keys = []

    async def asyncTearDown(self):
        for redis_key in self._redis_keys:
            await redis_key.delete()
        await super().asyncTearDown()

    async def test_simple_transaction(self):
        key = 'something'
        redis_list = self.add_redis_item(RedisList(self._redis, key))

        tx_result = []
        async with redis_list.begin_transaction() as transaction:
            transaction.add_operation(
                redis_list.enqueue('foo'),
                redis_list.get_range(),
                redis_list.enqueue('bar'),
                redis_list.length()
            )
            transaction.set_result_callback(lambda *args: tx_result.extend(args))

        result = await redis_list.get_range()

        self.assertEqual(result, ['bar', 'foo'])
        self.assertEqual(tx_result, [1, ['foo'], 2, 2])

    async def test_empty_transaction(self):
        key1 = 'something'
        redis_str = self.add_redis_item(RedisString(self._redis, key1))

        async with redis_str.begin_transaction():
            pass

    async def test_external_transaction(self):
        key1 = 'something'
        key2 = 'other-thing'
        redis_client = RedisClient(self._redis)
        redis_list = self.add_redis_item(RedisList(redis_client, key1))
        redis_str = self.add_redis_item(RedisString(redis_client, key2))

        async with redis_client.begin_transaction() as transaction:
            transaction.add_operation(
                redis_list.enqueue('test'),
                redis_str.set('special-value'),
                redis_list.enqueue('me')
            )

        items = await redis_list.get_range()
        value = await redis_str.get()

        self.assertEqual(items, ['me', 'test'])
        self.assertEqual(value, 'special-value')

    async def test_discard_transaction(self):
        key = 'something'
        redis_list = self.add_redis_item(RedisList(self._redis, key))

        try:
            async with redis_list.begin_transaction() as transaction:
                transaction.add_operation(
                    redis_list.enqueue('foo'),
                    redis_list.enqueue('bar')
                )
                raise Exception()
        except Exception:
            pass

        await redis_list.enqueue('outstanding', 'task')

        result = await redis_list.get_range()

        self.assertEqual(result, ['task', 'outstanding'])



from os import environ as env
import unittest
from aioredis import Redis, create_redis_pool

class RedisTests(unittest.IsolatedAsyncioTestCase):
    _redis: Redis = None

    async def asyncSetUp(self):
        self._redis = await create_redis_pool(env['REDIS_URL'])

    async def asyncTearDown(self):
        self._redis.close()
        await self._redis.wait_closed()



"""This module contains the following classes:
- RedisList: Represents a list stored in Redis.
"""

from functools import partial
from typing import List, Tuple, Any, Awaitable, AsyncIterator
from .redis_key import RedisKey
from .asyncio_utils import noop


class RedisList(RedisKey):
    """
    Represents a list store in Redis.
    """

    def length(self) -> Awaitable[int]:
        """
        Gets the length of the list.

        Returns:
            Awaitable[int]: The length of the list.
        """

        return self.get_connection().llen(self._key)

    def get_range(self, start: int=0, stop: int=-1, encoding='utf-8') -> Awaitable[List]:
        """
        Gets the given sub-sequence of the list.

        Args:
            start (int, optional): The start index of the range get. Defaults to 0.
            stop (int, optional): The stop index of the range to get. Negative indices are offsets
                from the length of the sequence. Defaults to -1.
            encoding (str, optional): The encoding to use for decoding the values. Defaults
                to 'utf-8'.

        Returns:
            Awaitable[List]: The retrieved range as a list.
        """

        return self.get_connection().lrange(self._key, start, stop, encoding=encoding)

    async def enumerate(
        self,
        start: int=0,
        stop: int=None,
        batch_size: int=0,
        encoding='utf-8'
    ) -> AsyncIterator[Any]:
        """
        Enumerates the items of this list in batches. This operation is not atomic and cannot be
        performed transactionally.

        Args:
            start (int, optional): The index to start from. Defaults to 0.
            stop (int, optional): The index to stop at. A value of None indicates no stop index.
                Defaults to None.
            batch_size (int, optional): The number of items to get in each batch.
                A value of 0 or None indicates a batch size equal to the full length of the list.
                Defaults to 0.
            encoding (str, optional): The encoding to use for the items. Defaults to 'utf-8'.

        Returns:
            AsyncIterator[Any]: An iterator that can be used to iterate over the result.
        """
        current_start = start
        while True:
            possible_stop = current_start + batch_size - 1
            current_stop = min(possible_stop if stop is None else stop, possible_stop) \
                if batch_size else (-1 if stop is None else stop)
            len_result = 0
            for item in await self.get_range(current_start, current_stop, encoding=encoding):
                len_result += 1
                yield item
            current_start = current_start + len_result
            if current_stop == -1 or \
                current_start <= current_stop or \
                (stop is not None and current_stop >= stop):
                break

    def push(self, *value: Tuple, reverse: bool=False) -> Awaitable[int]:
        """
        Pushes the given values into the list.

        Args:
            value (Tuple): The values to push into the list.
            reverse (bool, optional): Whether to push the values at the end of the list. Defaults
                to `False`.

        Returns:
            Awaitable[int]: The length of the list after the push operation.
        """

        value = list(filter(None, value))
        if not value:
            return noop()
        func = self.get_connection().rpush if reverse else self.get_connection().lpush
        return func(self._key, *value)

    def pop(
        self,
        reverse: bool=False,
        block: bool=False,
        timeout_seconds: int=0,
        encoding='utf-8'
    ) -> Awaitable[Any]:
        """
        Pops a value from the list.

        Args:
            reverse (bool, optional): Whether to pop the value from the end of
                the list. Defaults to `False`.
            block (bool, optional): Whether to block until an item is available to pop. Defaults to
                `False`.
            timeout_seconds (int, optional): The amount of time in seconds to wait before giving up.
                Defaults to 0, which indicates no timeout.
            encoding (str, optional): The encoding to use for decoding the popped value. Defaults
                to 'utf-8'.

        Returns:
            Awaitable[Any]: The value popped from the list, if any.
        """

        if reverse and block:
            func = partial(self.get_connection().brpop, timeout=timeout_seconds)
        elif reverse:
            func = self.get_connection().rpop
        elif block:
            func = partial(self.get_connection().blpop, timeout=timeout_seconds)
        else:
            func = self.get_connection().lpop

        return func(self._key, encoding=encoding)

    def enqueue(self, *value: Tuple) -> Awaitable[int]:
        """
        Enqueues the given values into the list.

        Args:
            value (Tuple): The values to enqueue.

        Returns:
            Awaitable[int]: The length of the list after the push operation.
        """

        return self.push(*value)

    def dequeue(
        self,
        block: bool=False,
        timeout_seconds: int=0,
        encoding='utf-8'
    ) -> Awaitable[Any]:
        """
        Dequeues an item from the list.

        Args:
            block (bool, optional): Whether to block until an item is available to dequeue.
                Defaults to `False`.
            timeout_seconds (int, optional): The amount of time in seconds to wait before giving
                up. Defaults to 0, which indicates no timeout.
            encoding (str, optional): The encoding to use for decoding the dequeued value.
                Defaults to 'utf-8'.

        Returns:
            Awaitable[Any]: The value dequeued from the list, if any.
        """

        return self.pop(
            reverse=True,
            block=block,
            timeout_seconds=timeout_seconds,
            encoding=encoding
        )

    def move(
        self,
        destination_key: str,
        block: bool=False,
        timeout_seconds: int=0,
        encoding='utf-8'
    ) -> Awaitable[Any]:
        """
        Moves a value from the end of one list to the beginning of another.

        Args:
            destination_key (str): The key of the list to move popped item to.
            block (bool, optional): Whether to block until an item is available to pop. Defaults to
                `False`.
            timeout_seconds (int, optional): The amount of time in seconds to wait before giving
                up. Defaults to 0.
            encoding (str, optional): The encoding to use for decoding the popped value. Defaults
                to 'utf-8'.

        Returns:
            Awaitable[Any]: The value popped from the list, if any.
        """

        func = partial(
            self.get_connection().brpoplpush,
            timeout=timeout_seconds
        ) if block else self.get_connection().rpoplpush
        return func(self._key, destination_key, encoding=encoding)

    def requeue(
        self,
        block: bool=False,
        timeout_seconds: int=0,
        encoding='utf-8'
    ) -> Awaitable[Any]:
        """
        Removes a value from the beginning of the list and pushes it to the end of the same list.

        Args:
            block (bool, optional): Whether to block until an item is available. Defaults to
                `False`.
            timeout_seconds (int, optional): The amount of time to wait before giving up. Defaults
                to 0.
            encoding (str, optional): The encoding to use for decoding the popped value. Defaults
                to 'utf-8'.

        Returns:
            Awaitable[Any]: The value popped from the list, if any.
        """

        return self.move(
            self._key,
            block=block,
            timeout_seconds=timeout_seconds,
            encoding=encoding
        )

    def remove(self, value: str, count: int=0) -> Awaitable[int]:
        """
        Removes occurrences of the given value from the list.

        Args:
            value (str): The value to remove.
            count (int, optional): The number of occurrences to remove. Defaults to 0, which
                removes all.

        Returns:
            Awaitable[int]: The number of items that were removed.
        """

        return self.get_connection().lrem(self._key, count, value)

    async def find_index(
        self,
        value: Any,
        start: int=0,
        stop: int=-1,
        batch_size: int=0,
        encoding='utf-8',
    ) -> int:
        """
        Finds the index of the given value, if any. This operation is not atomic and cannot be
        performed transactionally.

        Args:
            value (Any): The value to look for.
            start (int, optional): The index to start looking from. Defaults to 0.
            stop (int, optional): The index to stop at. A value of None indicates no stop index.
                Defaults to None.
            batch_size (int, optional): The number of items to get in each batch.
                A value of 0 or None indicates a batch size equal to the full length of the list.
                Defaults to 0.
            encoding (str, optional): The encoding to use for decoding values. Defaults to 'utf-8'.

        Returns:
            Awaitable[int]: The index of the provided value. `None` if not found.
        """

        index = start
        async for item in self.enumerate(
            start=start,
            stop=stop,
            batch_size=batch_size,
            encoding=encoding
        ):
            if item == value:
                return index
            index += 1
        return None



"""
This module contains the following classes:
- RedisString: Represents a string stored in Redis.
"""

from typing import Awaitable, Union
from .redis_key import RedisKey


class RedisString(RedisKey):
    """
    Represents a string stored in Redis.
    """

    def length(self) -> Awaitable[int]:
        """
        Gets the length of the string.

        Returns:
            Awaitable[int]: The length of the string.
        """

        return self.get_connection().strlen(self._key)

    def get(self, encoding='utf-8') -> Awaitable[str]:
        """
        Gets the stored value of the string.

        Returns:
            Awaitable[str]: The value of the string.
        """

        return self.get_connection().get(self._key, encoding=encoding)

    def set(
        self,
        value: Union[str, int, float],  # pylint:disable=unsubscriptable-object
        timeout_seconds: Union[int, float]=None,  # pylint:disable=unsubscriptable-object
        if_exists_equals: bool=None
    ) -> Awaitable:
        """
        Sets the string to a given value.

        Args:
            value (Union[str, int, float]): The value to set.
            timeout_seconds (Union[int, float], optional): The amount of time in seconds after which
                the key should expire. Defaults to `None`.
            if_exists_equals (bool, optional): If `True`, will only set this value if key already
                exists;
                if `False`, will only set this value if key does not already exist;
                if `None`, the value will always be set regardless of whether it already exists or
                not. Defaults to `None`.

        Returns:
            Awaitable: The result of the operation.
        """

        if if_exists_equals is True:
            exist = 'SET_IF_EXIST'
        elif if_exists_equals is False:
            exist = 'SET_IF_NOT_EXIST'
        else:
            exist = None
        return self.get_connection().set(
            self._key,
            value,
            pexpire=round(timeout_seconds * 1000) if timeout_seconds else None,
            exist=exist
        )



"""
This module contains the following classes:
- RedisSet: Represents a set stored in Redis.
"""

from typing import Awaitable, List
from .redis_key import RedisKey
from .asyncio_utils import noop


class RedisSet(RedisKey):
    """
    Represents a set stored in Redis.
    """

    def size(self) -> Awaitable[int]:
        """
        Gets the size of the set.

        Returns:
            Awaitable[int]: The size of the set.
        """

        return self.get_connection().scard(self._key)

    def get_all(self, encoding='utf-8') -> Awaitable[List]:
        """
        Gets all the members of the set.

        Args:
            encoding (str, optional): The encoding to use when decoding set members. Defaults to
                'utf-8'.

        Returns:
            Awaitable[List]: The members of the set.
        """

        return self.get_connection().smembers(self._key, encoding=encoding)

    def add(self, value: str) -> Awaitable[int]:
        """
        Adds an item to the set.

        Args:
            value (str): The item to add.

        Returns:
            Awaitable[int]: The number of items that were added to the set.
        """

        return noop() if value is None else self.get_connection().sadd(self._key, value)

    def remove(self, value: str) -> Awaitable[int]:
        """
        Removes an item from the set.

        Args:
            value (str): The item to remove.

        Returns:
            Awaitable[int]: The number of elements that were removed from the set.
        """

        return self.get_connection().srem(self._key, value)



"""
This module contains the following classes:
- RedisDoubleHash: Represents a two-way hash map stored in Redis.
"""

from typing import Awaitable, List, Set, Union
# Aliasing this to avoid confusion with the `set` function below.
from builtins import set as builtin_set
from aioredis import Redis
from .redis_model import RedisModel
from .redis_client import RedisClient
from .redis_key import RedisKey
from .redis_set import RedisSet


class RedisDoubleHash(RedisModel):
    """
    Represents a two-way hash map stored in Redis. Each field can be associated with multiple
    values and each value can be associated with multiple fields. The structure allows for
    getting all the values for a field and all the fields for a given value. The values are
    thus referred to as inverted fields. Since this class manages more than one Redis structure,
    some of the methods here are not completely atomic and most of them cannot participate in
    transactions.
    """

    def __init__(
        self,
        redis: Union[Redis, RedisClient],  # pylint:disable=unsubscriptable-object
        key: str, inverse_key: str
    ):
        """
        Creates an instance of `RedisDoubleHash`.

        Args:
            redis (Redis): The Redis instance to use to connect to Redis.
            key (str): The key to use for forward hash map.
            inverse_key (str): The key to use for inverted hash map.
        """

        super().__init__(redis)
        self._key = key
        self._inverse_key = inverse_key

    async def fields(self) -> Awaitable[Set]:
        """
        Gets all the fields in the hash map.

        Returns:
            Awaitable[Set]: The fields in the hash map.
        """

        return builtin_set(
            self._extract_field_name(redis_key, self._key) \
                for redis_key in await self._fields_generic()
        )

    async def fields_inverted(self) -> Awaitable[Set]:
        """
        Gets all the inverted fields (values) in the hash map.

        Returns:
            Awaitable[Set]: The inverted fields (values) in the hash map.
        """

        return builtin_set(
            self._extract_field_name(redis_key, self._inverse_key) \
                for redis_key in await self._fields_generic(inverse=True)
        )

    def get(self, field: str) -> Awaitable[List]:
        """
        Gets the inverted fields associated with the given field.

        Args:
            field (str): The field to get.

        Returns:
            Awaitable[List]: The list of all inverted fields associated with the given field.
        """

        return self._get_field_value(self._key, field)

    def get_inverted(self, field: str) -> Awaitable[List]:
        """
        Gets the fields associated with the given inverted field (value).

        Args:
            field (str): The inverted field to get.

        Returns:
            Awaitable[List]: The list of all fields associated with the given inverted field.
        """

        return self._get_field_value(self._inverse_key, field)

    async def set(self, field: str, value: str):
        """
        Associates the given value with the given field.

        Args:
            field (str): The name of the field.
            value (str): The value to associate.
        """

        if value is None:
            return

        async with self.begin_transaction() as transaction:
            transaction.add_operation(
                self._set_field(self._key, field, value),
                self._set_field(self._inverse_key, value, field)
            )

    async def unset(self, field: str, value: str):
        """
        Dissociates the given value with the given field.

        Args:
            field (str): The name of the field.
            value (str): The name of the value to dissociate.
        """

        if value is None:
            return

        async with self.begin_transaction() as transaction:
            transaction.add_operation(
                self._unset_field(self._key, field, value),
                self._unset_field(self._inverse_key, value, field)
            )

    async def set_inverted(self, field: str, value: str):
        """
        Associates the given value with the inverted field.

        Args:
            field (str): The name of the inverted field.
            value (str): The value to associate.
        """

        if value is None:
            return
        return await self.set(field=value, value=field)

    async def remove(self, field: str):
        """
        Removes the given field from both sides of the hash map.

        Args:
            field (str): The field to remove.
        """

        return await self._remove_generic(self._key, self._inverse_key, field)

    async def remove_inverted(self, field: str):
        """
        Removes the given inverted field from both sides of the hash map.

        Args:
            field (str): The inverted field to remove.
        """

        return await self._remove_generic(self._inverse_key, self._key, field)

    async def delete(self):
        """
        Deletes all mappings from both sides of the hash map.
        """

        all_keys = builtin_set()

        async with self.begin_transaction() as transaction:
            transaction.add_operation(
                self._fields_generic(),
                self._fields_generic(inverse=True)
            )
            transaction.set_result_callback(
                lambda *result: all_keys.update(
                    key for keys in result for key in keys
                )
            )

        async with self.begin_transaction() as transaction:
            transaction.add_operation(*(
                RedisKey(self.get_connection(), key).delete() for key in all_keys
            ))

    def _fields_generic(self, inverse: bool=False) -> Awaitable[List]:
        return self.get_connection().keys(self._get_field_name(
            self._inverse_key if inverse else self._key,
            '*'
        ), encoding='utf-8')

    def _get_field_value(self, key: str, field: str) -> Awaitable[List]:
        sub_set = self._get_redis_set(key, field)
        return sub_set.get_all()

    def _set_field(self, key: str, field: str, value: str):
        sub_set = self._get_redis_set(key, field)
        return sub_set.add(value)

    def _unset_field(self, key: str, field: str, value: str):
        sub_set = self._get_redis_set(key, field)
        return sub_set.remove(value)

    async def _remove_generic(self, key: str, inverse_key: str, field: str):
        sub_set = self._get_redis_set(key, field)
        values = await sub_set.get_all()
        async with self.begin_transaction() as transaction:
            transaction.add_operation(sub_set.delete())
            for value in values:
                value_set = RedisSet(self._redis, self._get_field_name(inverse_key, value))
                transaction.add_operation(value_set.remove(field))

    def _get_redis_set(self, key: str, field: str) -> RedisSet:
        return RedisSet(self._redis, self._get_field_name(key, field))

    @staticmethod
    def _get_field_name(key: str, field: str) -> str:
        return f'{key}:{field}'

    @staticmethod
    def _extract_field_name(redis_key: str, key: str) -> str:
        return redis_key.removeprefix(key + ':')



"""
This module contains the following classes:
- RedisTransaction: Represents a Redis transaction.
"""

from typing import Any, Callable, List, Tuple
from asyncio import Future, gather


class RedisTransaction:
    """
    Represents a Redis transaction that can also be used as a context manager.
    """

    _tasks: List[Future]
    _result_callback: Callable[[Tuple[Any, ...]], Any]

    def __init__(self, redis_client):
        """
        Creates an instance of `RedisTransaction`.

        Args:
            redis_client (RedisClient): The instance to use for connecting to Redis.
        """
        self._redis_client = redis_client
        self._tasks = []
        self._result_callback = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, traceback):
        if not exc:
            await self._redis_client.execute_transaction()
            result = await gather(*self._tasks)
            if self._result_callback:
                self._result_callback(*result)
        else:
            self._redis_client.discard_transaction()
            for task in self._tasks:
                task.cancel()

    def add_operation(self, *tasks: Future):
        """
        Adds the given operation(s) to the transaction.

        Args:
            tasks (tuple[Future, ...]): The tasks to add to the transaction. Should be Redis
            operations that support transactions.
        """
        self._tasks.extend(tasks)

    def set_result_callback(self, callback: Callable[[Tuple[Any, ...]], Any]):
        """
        Sets a callback function to be called with the result of the transaction when it completes.

        Args:
            callback (Callable[[Tuple[Any, ...]], Any]): The function to call. Accepts positional
                arguments containing results of operations in the same order as they were added
                to the transaction using calls to `add_operation`.
        """
        self._result_callback = callback
